@@ -535,18 +535,28 @@ namespace types
       return (*this)[std::get<0>(indices)][tuple_tail(indices)];
     }
 
-    template <class Ty0, class Ty1, class... Tys>
-    auto operator[](std::tuple<Ty0, Ty1, Tys...> const &indices) const ->
-        typename std::enable_if<is_numexpr_arg<Ty0>::value,
-                                decltype((*this)[tuple_tail(indices)])>::type;
-
     template <class Slices, size_t... Is>
     auto _fwdindex(Slices const &indices, utils::index_sequence<Is...>) const
         & -> decltype((*this)(std::get<Is>(indices)...))
+    {
+      return (*this)(std::get<Is>(indices)...);
+    }
+    template <class S, size_t... Is>
+    auto _fwdindex(dynamic_tuple<S> const &indices,
+                   utils::index_sequence<Is...>) const
+        & -> decltype((*this)(std::get<Is>(indices)...))
     {
       return (*this)((indices.size() > Is ? std::get<Is>(indices)
                                           : contiguous_slice())...);
     }
+
+    template <class Ty0, class Ty1, class... Tys>
+    auto operator[](std::tuple<Ty0, Ty1, Tys...> const &indices) const ->
+        typename std::enable_if<is_numexpr_arg<Ty0>::value,
+                                decltype(this->_fwdindex(
+                                    indices, utils::make_index_sequence<
+                                                 2 + sizeof...(Tys)>()))>::type;
+
     template <class Ty, size_t M>
     auto operator[](array<Ty, M> const &indices) const & ->
         typename std::enable_if<
@@ -697,7 +707,9 @@ namespace types
     };
 #if defined(_WIN32) || defined(__APPLE__)
     template <>
-    struct dtype_helper<long> : dtype_helper<int64_t> {
+    struct dtype_helper<long>
+        : dtype_helper<typename std::conditional<
+              sizeof(long) == sizeof(int64_t), int64_t, int32_t>::type> {
     };
 #endif
     template <>
@@ -888,7 +900,9 @@ PYTHONIC_NS_END
 template <class T1, class T2, class pS1, class pS2>
 struct __combined<pythonic::types::ndarray<T1, pS1>,
                   pythonic::types::ndarray<T2, pS2>> {
-  using type = pythonic::types::ndarray<typename __combined<T1, T2>::type, pS1>;
+  using type = pythonic::types::ndarray<
+      typename __combined<T1, T2>::type,
+      pythonic::sutils::common_shapes_t<std::tuple_size<pS1>::value, pS1, pS2>>;
 };
 
 template <class pS, class T, class O>
